#ifndef __SERVER_H__
#define __SERVER_H__

#include <util/string.h>
#include <util/handler.h>
#include <util/locker.h>
#include <map>

#include "transferthread.h"
#include "queuenode.h"


using namespace os;

/** \brief Server class.
 * This is the base server class that provides a common interface for
 * accessing various types of servers. This should be sub-classed for
 * the different types of servers.
 *
 * \todo ParseDirListing function should be removed if it is unnecessary.
 */
class Server
{
public:
	Server( const String& zURL, Handler* pcTarget, int nParallelConnections = 2 );
	virtual ~Server();
	int Close();
	
	int Lock();
	int Unlock();
	
	String GetServerAddress();
	
	int GetDirListing( const String& zPath, Handler* pcTarger );
	int GetRemote( const String& zRemoteSource, const String& zLocalDest );
	int SendLocal( const String& zLocalSource, const String& zRemoteDest );

#if 0
	virtual int ParseDirListing( const String& zDir, char* pzBuf, size_t nSize, Handler* pcTarget, bool bInitial );	/* Parse the text of a dir listing into RemoteNodes and send it to the RemoteView */	
#endif

	/* Methods for controlling active transfers, eg pause, cancel, resume */
	int PauseTransfer( int nNodeID );
	int ResumeTransfer( int nNodeID );
	int CancelTransfer( int nNodeID );
	
	/* Methods for manipulating the queue (ie raise/lower priority of a queued transfer) */
	/* TODO */
	
private:
	/* These methods are only called from TransferThread */

	virtual CURLcode _SetupCurlHandle( CURL* pHandle );	/* Add our settings to the curl handle */

private:

	String m_zURL; /**< Address of the server. */

	Handler* m_pcTarget; /**< The window that created us. */
	
	TransferThread* m_pcThread;	/**< Our transfer thread, that does all the work. */

	int m_nParallelConnections;	/**< Number of concurrent connections (ie curl handles) to use. */
	
	
	/** \brief List of all queued and running transfers stored as a map from id numbers to QueueNode objects.
	 * The reason for using a std::map and giving each node an id number is so that the gui can refer to queue nodes
	 * by their id rather than using a pointer. Since we can't be sure that the transfer thread hasn't deleted a node
	 * while the gui is still using the pointer.
	 */
	std::map< int, QueueNode* > m_apcQueue;
	
	
	int m_nNextID; /**< Next available ID number. */
	Locker* m_pcLock; /**< Lock used for concurrency. */
	
	friend class TransferThread;
};

/** \brief FTP Server
 * This is an instance of the Server class that implements the
 * FTP Protocols.
 *
 * \todo Functionality specific for FTP style connections should be moved here
 * in order to better abstract the Server object.
 */
class FTPServer : public Server
{
public:
	FTPServer( const String& zURL, Handler* pcTarget, int nParallelConnections );

//	int ParseDirListing( const String& zDir, char* pzBuf, size_t nSize, Handler* pcTarget, bool bInitial );
	
private:
	~FTPServer();
//	CURLcode _SetupCurlHandle( CURL* pHandle );
};



#endif	/* __SERVER_H__ */




