/* TODO:
	- cleanup
*/

#include "transferthread.h"
#include "queuenode.h"
#include "server.h"
#include "messages.h"

#include <util/message.h>
#include <util/exceptions.h>

#include <unistd.h>
#include <sys/select.h>

using namespace os;

#define max( a, b ) (a > b ? a : b)

/** \brief TransferThread Constructor.
 * Initializes the TransferThread creating its pipes, locks and curl objects.
 * \param pcServer The server that this thread is working for.
 */
TransferThread::TransferThread( Server* pcServer ) : Thread( "transfer_thread" )
{
	m_pcServer = pcServer;
	
	m_pcLock = new Locker( "transfer_thread_lock" );
	
	int nTmp[2];
	int nResult;
	nResult = pipe( nTmp );		/* Create a pipe for communication between transfer_thread and others */
	if( nResult ) {
		printf( "TransferThread: Unable to create pipe!\n" );
		throw( errno_exception( "TransferThread: Unable to create pipe!" ) );
	}
	m_nReadPipe = nTmp[0];
	m_nWritePipe = nTmp[1];
	
	/* Set the pipe to non-blocking IO */
	fcntl( m_nReadPipe, F_SETFL, (fcntl( m_nReadPipe, F_GETFL ) | O_NONBLOCK) );
	fcntl( m_nWritePipe, F_SETFL, (fcntl( m_nWritePipe, F_GETFL ) | O_NONBLOCK) );
	
	CURL* pHandle = curl_multi_init();
	if( pHandle == NULL ) {
		printf( "Transferrer: Could not create curl multi handle!\n" );
		throw( errno_exception( "Could not create curl multi handle!\n" ) );
	}
	m_pCurlMultiHandle = pHandle;
	
	for( int i = 0; i < m_pcServer->m_nParallelConnections; i++ ) {
		CURL* pHandle = curl_easy_init();
		if( m_pcServer->_SetupCurlHandle( pHandle ) != CURLE_OK ) {
			printf( "Transferrer: Could not setup curl handle!\n" );
			throw( errno_exception( "Could not setup curl handle!" ) );
		}
		CurlHandle_s* psTmp = new CurlHandle_s;		
		psTmp->m_pHandle = pHandle;
		psTmp->m_nStatus = HANDLE_IDLE;
		curl_easy_setopt( pHandle, CURLOPT_PRIVATE, (void*)psTmp );		/* Save the CurlHandle_s to the curl handle's private data */
		
		m_asHandles.push_back( psTmp );
	}
}

/** \brief TransferThread Deconstructor.
 * This deletes the lock for this thread and cleans up curl data.
 *
 * \todo Other curl cleanup - remove all easy handles from the multi handle
 */
TransferThread::~TransferThread()
{
	delete( m_pcLock );
	
	curl_multi_cleanup( m_pCurlMultiHandle );
}

/** \brief Lock the Thread.
 * This is used to lock the thread from further access until any operations
 * being performed are complete.
 */
int TransferThread::Lock()
{
	return( m_pcLock->Lock() );
}

/** \brief Unlock the Thread.
 * This unlocks the thread for further operations.
 */
int TransferThread::Unlock()
{
	return( m_pcLock->Unlock() );
}

/** \brief Wrapper for sending messages with numerical codes.
 * This creates a Message object using nCode as a parameter.
 *
 * \param nCode The code to send for the message.
 */
int TransferThread::SendMessage( int nCode )
{
	return( SendMessage( new Message( nCode ) ) );
}

/** \brief Send a message to the thread.
 * The message will be deleted by the thread once it is processed,
 * so it should be created via new Message( code ).
 *
 * \warning The messages added to m_apcMessages by this function should be deleted after processing.
 */
int TransferThread::SendMessage( Message* pcMsg )
{
	Lock();
	m_apcMessages.push( pcMsg );
	Unlock();
	_Notify();
	return( 0 );
}

/** \brief Notify() writes a piece of data to the pipe to wake the transfer 
 * thread from its select() sleep. This is called from gui threads.
 */
void TransferThread::_Notify()
{
	if( get_thread_id( NULL ) == GetThreadId() ) {
		printf( "Warning: TransferThread::_Notify() called from transfer thread!\n" );
		return;
	}	
	write( m_nWritePipe, "x", 1 );		/* This will wake the transfer_thread from its select() sleep */
}

/** \brief _CleanupNode() cleans up and deletes a finished QueueNode.
 *
 * \param nNode The QueueNode ID to remove.
 *
 * \todo Notify GUI.
 */
void TransferThread::_CleanupNode( int nNode )
{
	m_pcServer->Lock();
	if( m_pcServer->m_apcQueue.count( nNode ) == 0 ) {
		/* Sanity checking - better safe than sorry */
		printf( "_CleanupNode( %i ): node %i doesn't exist or has already been cleaned up!!\n", nNode, nNode );
		m_pcServer->Unlock();
		return;
	}
	QueueNode* pcNode = m_pcServer->m_apcQueue[ nNode ];
	m_pcServer->m_apcQueue.erase( nNode );
	delete( pcNode );
	
	m_pcServer->Unlock();
}

/** \brief Assign queued transfers to idle curl handles.
 * This schedules a transfer for one of the curl handles.
 */
void TransferThread::_Schedule()
{
	for( uint i = 0; i < m_asHandles.size(); i++ )
	{
		CurlHandle_s* psHandle = m_asHandles[i];
		if( psHandle->m_nStatus == HANDLE_IDLE )
		{
			printf( "TransferThread::_Schedule(): curl handle %x is free\n", psHandle->m_pHandle );
			/* This curl handle is idle. Find the next queued node from the queue and setup the curl handle */
			m_pcServer->Lock();
			std::map< int, QueueNode* >::iterator j = m_pcServer->m_apcQueue.begin();
			for( ; j != m_pcServer->m_apcQueue.end(); j++ )
			{
				QueueNode* pcNode = j->second;
				if(pcNode->m_nStatus == NODE_QUEUED )
				{
					printf( " _Schedule(): assigning %s node %s to handle %x\n", (pcNode->m_nType == NODE_DOWNLOAD ? "download": (pcNode->m_nType == NODE_UPLOAD ? "upload" : "dirlist")), pcNode->m_zLocalPath.c_str(), psHandle->m_pHandle );
					CURLcode nResult;
					nResult = pcNode->SetupCurlHandleForTransfer( psHandle->m_pHandle );
					if( nResult != CURLE_OK ) {
						psHandle->m_nStatus = HANDLE_INVALID;
						continue;
					}
					
					CURLMcode nResult2;
					nResult2 = curl_multi_add_handle( m_pCurlMultiHandle, psHandle->m_pHandle );
					if( nResult2 != CURLM_OK ) {
						printf( "TransferThread: curl_multi_add_handle() failed: %s (%i)\n", curl_multi_strerror( nResult2 ), nResult2 );
						continue;
					}
					pcNode->m_nStatus = NODE_RUNNING;
					psHandle->m_nStatus = pcNode->m_nID;
					break;
				}
			}
			if( j == m_pcServer->m_apcQueue.end() ) printf( " _Schedule(): handle %x is idle (queue is empty).\n", psHandle->m_pHandle );
			m_pcServer->Unlock();
		}
	}
}


/** \brief Checks for messages from other threads and carries out the appropriate 
 * action (adding handles to curl_multi_handle, etc).
 *
 * \warning Only called from the transfer thread.
*/
void TransferThread::_ProcessMessages()
{
	Lock();
	while( !m_apcMessages.empty() ) {

		Message* pcMsg = m_apcMessages.front();
		m_apcMessages.pop();

		
		switch( pcMsg->GetCode() )
		{
			case M_TT_CLOSE:
			{
				/* TODO: Notify gui somewhere? */
				/* Clean up all transfers, free curl handles and delete the Server then ourself */
				printf( "TransferThread (%i): M_TT_CLOSE; performing cleanup\n", get_thread_id( NULL ) );
				
				/* Close curl handles and delete CurlHandle_s structs */
				for( int i = 0; i < m_asHandles.size(); i++ ) {
					CurlHandle_s* pcHandle = m_asHandles[i];
					curl_multi_remove_handle( m_pCurlMultiHandle, pcHandle->m_pHandle );
					curl_easy_cleanup( pcHandle->m_pHandle );
					delete( pcHandle );
				}
				curl_multi_cleanup( m_pCurlMultiHandle );
				printf( "  curl handles deleted.\n" );

				m_pcServer->Lock();
				/* Delete all nodes from the queue */
				std::map< int, QueueNode* >::iterator i = m_pcServer->m_apcQueue.begin();
				for( ; i != m_pcServer->m_apcQueue.end(); i++ )
				{
					_CleanupNode( i->first );
				}
				m_pcServer->Unlock();
				printf( "  queue nodes deleted.\n" );

				/* Delete the Server object */
				delete( m_pcServer );
				printf( "  Server deleted.\n" );
				
				/* Delete any remaining messages */
				m_pcLock->Lock();
				delete( pcMsg );
				while( !m_apcMessages.empty() ) {
					Message* pcTmp = m_apcMessages.front();
					m_apcMessages.pop();
					delete( pcTmp );
				}
				m_pcLock->Unlock();
				delete( m_pcLock );
				printf( "  messages deleted.\n" );

				/* Close the pipe */
				close( m_nReadPipe );
				close( m_nWritePipe );
				printf( "  pipes closed; about to Terminate().\n" );
				fflush( stdout );		/* Because we get some wierdness in the terminal when two threads are printf'ing at the same time */
				
#if 0	/* Not working - causes heap corruption */
				/* Delete ourself and quit the thread */
				delete( this );		/*	NOTE: I don't think this is strictly kosher with the current libsyllable.
										The destructor calls Terminate() which cals kill( our thread, SIGKILL ).
										So the destructor never returns and the TransferThread object's memory isn't freed.
										So the next line currently isn't reached, and the TransferThread memory is leaked.
										This probably should be fixed in libsyllable. When it is, hopefully this code will be correct.
									*/
				printf( "  delete( this ) done!\n" );
				exit_thread( 0 );
#endif
				Terminate();
				break;
			}
			case M_TT_ADD:
			{
				String zLocalPath, zRemotePath;
				int nType;		/* Download, upload, dirlist */
				if( pcMsg->FindString( "remotepath", &zRemotePath ) || pcMsg->FindString( "localpath", &zLocalPath ) || pcMsg->FindInt32( "type", &nType ) ) {
					printf( "TransferThread: Got M_TT_ADD without necessary parameters!\n" );
					break;
				}
				
				/* TODO: Should move this code into a _AddNode() method */
				bool bFound = false;
				m_pcServer->Lock();
				std::map< int, QueueNode* >::iterator i = m_pcServer->m_apcQueue.begin();

				/* Check that there isn't already a queued transfer with this source and destination */
				for( ; i != m_pcServer->m_apcQueue.end(); i++ ) {
					QueueNode* pcNode = i->second;
					if( pcNode->m_zLocalPath == zLocalPath && pcNode->m_zRemotePath == zRemotePath && pcNode->m_nType == nType ) {
						printf( "TransferThread: Got duplicate M_TT_ADD for %s %s %s\n", zRemotePath.c_str(), (nType == NODE_UPLOAD?"<-":(nType == NODE_DOWNLOAD?"->":"<dir>")), zLocalPath.c_str() );
						bFound = true;
						break;
					}
				}
				if( bFound ) {
					m_pcServer->Unlock();
					break;
				}
				
				/* Add the transfer to the queue */
				QueueNode* pcTmp;
				int nID = m_pcServer->m_nNextID++;
				if( nType == NODE_DOWNLOAD ) {
					pcTmp = new DownloadNode( m_pcServer, zLocalPath, zRemotePath, nID );
				} else if( nType == NODE_UPLOAD ) {
					pcTmp = new UploadNode( m_pcServer, zLocalPath, zRemotePath, nID );
				} else if( nType == NODE_DIRLIST ) {
					Handler* pcTarget;
					if( pcMsg->FindPointer( "target", (void**)&pcTarget ) != 0 ) {
						printf( "TransferThread: Got M_ADD dirlist without target Handler!\n" );
						break;
					}
					pcTmp = new DirListNode( m_pcServer, zRemotePath, pcTarget, nID );
				} else {
					printf( "TransferThread: Got M_ADD with unknown type!\n" );
					break;
				}

				pcTmp->m_nStatus = NODE_QUEUED;
				m_pcServer->m_apcQueue[ nID ] = pcTmp;

				m_pcServer->Unlock();
				/* Schedule, in case there is a free handle and we can start this transfer immediately */
				_Schedule();
				break;
			}
			case M_TT_PAUSE:
			{
				/* TODO:
					- check if the node is already attached to a curl handle, and if so, remove the curl handle & set it to idle
					- set the node's status to paused
					- send message to gui to notify it
				*/
				break;
			}
			case M_TT_RESUME:
			{
				break;
			}
			case M_TT_CANCEL:
			{
				break;
			}
			default:
			{
				printf( "_ProcessMessages(): Got unknown message code %i!\n", pcMsg->GetCode() );
				break;
			}
		}
		delete( pcMsg );
	}
	Unlock();
}

/** \brief The main loop of the thread.
 */
int32 TransferThread::Run()
{
	printf( "Transfer thread: TID %i\n", GetThreadId() );
	fd_set sReadSet, sWriteSet, sExcSet;
	struct timeval sSelectTime;
	int nMaxFD;

	CURLMcode nResult;
	int nRunning;

	/* TODO: if there are no transfers active, then call select only on our port with infinite timeout to save on unneeded calls to _ProcessMessages() and curl_multi_process() */
	while( 1 ) {
		FD_ZERO( &sReadSet );
		FD_ZERO( &sWriteSet );
		FD_ZERO( &sExcSet );

		/* Get curl's file descriptors for later call to select() */
		nResult = curl_multi_fdset( m_pCurlMultiHandle, &sReadSet, &sWriteSet, &sExcSet, &nMaxFD );
		if( nResult != CURLM_OK ) {
			printf( "TransferThread: curl_multi_fdset returned %i (%s)!\n", nResult, curl_multi_strerror( nResult ) );
			/* continue ? */
		}
		FD_SET( m_nReadPipe, &sReadSet );
		nMaxFD = max( nMaxFD, m_nReadPipe );
		
		/* Get curl's preferred timeout */
		long nTimeout;
		nResult = curl_multi_timeout( m_pCurlMultiHandle, &nTimeout );
		if( nResult != CURLM_OK ) {
			printf( "TransferThread: curl_multi_timeout returned %i (%s)!\n", nResult, curl_multi_strerror( nResult ) );
		}
//		printf( "curl_multi_timeout() requests timeout %lu\n", nTimeout );
		if( nTimeout == -1 ) nTimeout = 1000000;	/* curl doesn't have a preference; default timeout 1 second */
		if( nTimeout > 1000000 ) nTimeout = 1000000;	/* max 1 second timeout */
		sSelectTime.tv_sec = nTimeout / 1000000;
		sSelectTime.tv_usec = nTimeout % 1000000;
		
		/* Wait for some data to arrive or another thread to message us, via select() */
		int nCount;
//		printf( "About to select(); timeout %lu\n", (long)(sSelectTime.tv_sec*1000000 + sSelectTime.tv_usec) );
		nCount = select( nMaxFD + 1, &sReadSet, &sWriteSet, &sExcSet, &sSelectTime );
//		printf( "select() done; result %i\n", nCount );

		/* Clear the data from the pipe */
		char zBuf[16];
		while( read( m_nReadPipe, zBuf, 16 ) > 0 ) ;

		/* Check for messages from other threads */
		_ProcessMessages();

		/* Call curl_multi_perform() to transfer some data until it is happy. */
		/* Curl will call our callback functions Transfer::ReadCallback() etc to read data etc. */
//		printf( "Calling curl_multi_perform()\n" );
		while( curl_multi_perform( m_pCurlMultiHandle, &nRunning ) == CURLM_CALL_MULTI_PERFORM ) {
			printf( "%i transfers still running.\n", nRunning );
		}
		
		/* Get any responses from libcurl */
		CURLMsg* psMsg;
		int nRemaining;
		bool bNeedSchedule = false;
		while( (psMsg = curl_multi_info_read( m_pCurlMultiHandle, &nRemaining )) != NULL )
		{
			printf( "curl_multi_info_read(): code %i, handle %x, result %i, %i messages remaining\n", psMsg->msg, psMsg->easy_handle, psMsg->data.result, nRemaining );
			if( psMsg->msg == CURLMSG_DONE )
			{
				CurlHandle_s* pcHandle;
				curl_easy_getinfo( psMsg->easy_handle, CURLINFO_PRIVATE, (char**)&pcHandle );	/* Curl docs say to pass the pointer as char** */
				int nOldNode = pcHandle->m_nStatus;
				CURLMcode nResult;
				nResult = curl_multi_remove_handle( m_pCurlMultiHandle, pcHandle->m_pHandle );
				if( nResult != CURLM_OK ) {
					printf( "curl_multi_remove_handle() failed for handle %x! (%i) %s\n", pcHandle->m_pHandle, nResult, curl_multi_strerror( nResult ) );
				}
				pcHandle->m_nStatus = HANDLE_IDLE;
				bNeedSchedule = true;
				
				long nCode;
				curl_easy_getinfo( psMsg->easy_handle, CURLINFO_RESPONSE_CODE, &nCode );
				printf( " handle %x reports response %li\n", psMsg->easy_handle, nCode );
				
				_CleanupNode( nOldNode );	
				/* TODO: Notify the gui to update; call QueueNode::Finished() method which flushes data to disk etc */
			} else { printf( "Got unknown message code %i from curl_multi_info_read()!\n", psMsg->msg ); }
		}
		if( bNeedSchedule ) {
			_Schedule();
		}
	}

	return( 0 );
}


